/* bmp.cpp: Simple BMP file loader
 *
 * Function to load a BMP file etc.
 *
 * Ankit Mohan, 3/2/2003 - Got out of my OLD game code
 * Ankit Mohan, 10/07/2003 - Removed text/font related functions.
 * Ankit Mohan, 04/12/2004 - corrected bmp read and write using multiple of 4 scanlines
 * Abhinav Dayal, 04/13/2004 - Added load_padded_bmp function
 */

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <string.h>
#include <math.h>
#include "bmp.h"

/* from some windows header */
struct BITMAPFILEHEADER { 
	unsigned long	bfSize; 
	unsigned short	bfReserved1; 
	unsigned short	bfReserved2; 
	unsigned long	bfOffBits; 
};

/* from some windows header */
struct BITMAPINFOHEADER{
	unsigned long  biSize; 
	unsigned long   biWidth; 
	unsigned long   biHeight; 
	unsigned short   biPlanes; 
	unsigned short   biBitCount; 
	unsigned long  biCompression; 
	unsigned long  biSizeImage; 
	unsigned long   biXPelsPerMeter; 
	unsigned long   biYPelsPerMeter; 
	unsigned long  biClrUsed; 
	unsigned long  biClrImportant; 
};

/* from some windows header */
struct RGBTRIPLE {
	unsigned char	rgbtBlue; 
	unsigned char	rgbtGreen; 
	unsigned char	rgbtRed; 
};

/* Loads a BMP file and returns it, with height and width.
 * Allocation for the file is done in this function. Memory should be 
 * freed by the caller. 
 * The image data is returned in a 1-dimensional array that contains the pixel
 * RGBA values one from the bottom up to the top row (each row from left to
 * right). The alpha value is set to 255. 
 *
 * returns -1 if file not found
 *         -2 if compressed BMP (cannot load)
 *         -3 if not 24 bit BMP (cannot load)
 *

 * Sample code:
 *
 * int w, h;
 * unsigned char *img;
 * loadBMP("filename.bmp", &img, &h, &w);
 *
 * glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, img);
 *
 */
int load_BMP(char *name, unsigned char **image, int *width, int *height)
{
	FILE *bmpFile;
	BITMAPFILEHEADER bmpFileHeader;
	BITMAPINFOHEADER bmpInfoHeader;
	struct RGBTRIPLE rgb;
	unsigned long i, j, offset;
	unsigned short type;
	int padding;
	unsigned char scratch[10];

	bmpFile = fopen(name, "rb");
	if (!bmpFile)
		return -1;

	/* first read the first two bytes (type) */
	fread(&type, 2, 1, bmpFile);

	fread(&bmpFileHeader, sizeof(bmpFileHeader), 1, bmpFile); /* Read File Header */
	fread(&bmpInfoHeader, sizeof(bmpInfoHeader), 1, bmpFile); /* Read Info Header */
    
    if (bmpInfoHeader.biCompression)
		return -2; /* compression used */

    if (bmpInfoHeader.biBitCount != 24)
		return -3; /* not 24 bpp */

    *image = (unsigned char *)malloc(bmpInfoHeader.biWidth*bmpInfoHeader.biHeight*4);
    *height = bmpInfoHeader.biHeight;
    *width = bmpInfoHeader.biWidth;
    offset=0;

	/* number of bytes in each line of a bmp file is a multiple of 4.
	 * provide adequate padding
	 */
	padding = (4 - (*width * 3) % 4) % 4;


    for (j=0; j<bmpInfoHeader.biHeight; j++) {
		for (i=0; i<bmpInfoHeader.biWidth; i++) {
			fread(&rgb, sizeof(rgb), 1, bmpFile);

			(*image)[offset] = rgb.rgbtRed;
	        offset++;

		    (*image)[offset] = rgb.rgbtGreen;
	        offset++;

		    (*image)[offset] = rgb.rgbtBlue;
			offset++;

			(*image)[offset] = 255;
		    offset++;
		}
		fread(scratch, padding, 1, bmpFile); /* read the padding and ignore it */
	}

	fclose(bmpFile);

	return 0;
}



/* Save a BMP file.
 *  
 * The image data is expected in a 1-dimensional array that contains the pixel
 * RGBA values one from the bottom up to the top row (each row from left to
 * right). The alpha value is ignored. 
 *
 * returns -1 cannot write to file
 *          0 on success
 */
int save_BMP(char *name, unsigned char *image, int width, int height)
{
	FILE *bmpFile;
	BITMAPFILEHEADER bmpFileHeader;
	BITMAPINFOHEADER bmpInfoHeader;
	struct RGBTRIPLE rgb;
	unsigned long i, j, offset;
	int padding;
	unsigned char scratch[10];

	memset(scratch, 0, sizeof(scratch));

	bmpFile = fopen(name, "wb");
	if (!bmpFile)
		return -1;

	/* first write the first two bytes (type) */
	fwrite("BM", 2, 1, bmpFile);

	/* File header */
	memset(&bmpFileHeader, 0, sizeof(struct BITMAPFILEHEADER));
	bmpFileHeader.bfSize = 2 + sizeof(struct BITMAPFILEHEADER) + sizeof(struct BITMAPINFOHEADER)
			+ width * height * 3;
	bmpFileHeader.bfOffBits = 2 + sizeof(struct BITMAPFILEHEADER) + sizeof(struct BITMAPINFOHEADER);
	fwrite(&bmpFileHeader, sizeof(bmpFileHeader), 1, bmpFile); /* write it! */

	/* Info header */
	memset(&bmpInfoHeader, 0, sizeof(struct BITMAPINFOHEADER));
	bmpInfoHeader.biSize = 0x28;
	bmpInfoHeader.biWidth = width;
	bmpInfoHeader.biHeight = height;
	bmpInfoHeader.biPlanes = 1;
	bmpInfoHeader.biBitCount = 24;
	bmpInfoHeader.biCompression = 0;
	bmpInfoHeader.biSizeImage = width * height * 3;
	bmpInfoHeader.biXPelsPerMeter = 1;
	bmpInfoHeader.biYPelsPerMeter = 1;
	bmpInfoHeader.biClrUsed = 0;
	bmpInfoHeader.biClrImportant = 0;
	fwrite(&bmpInfoHeader, sizeof(bmpInfoHeader), 1, bmpFile); /* write it! */

    offset=0;

	/* number of bytes in each line of a bmp file is a multiple of 4.
	 * provide adequate padding
	 */
	padding = (4 - (width * 3) % 4) % 4;

    for (j=0; j<bmpInfoHeader.biHeight; j++) {
		for (i=0; i<bmpInfoHeader.biWidth; i++) {
			rgb.rgbtRed = image[offset];
	        offset++;
		    
			rgb.rgbtGreen = image[offset];
	        offset++;

		    rgb.rgbtBlue = image[offset];
			offset++;

		    offset++; /* ignore alpha */

			fwrite(&rgb, sizeof(rgb), 1, bmpFile);
		}
		fwrite(scratch, padding, 1, bmpFile); /* write the padding */
	}

	fclose(bmpFile);

	return 0;
}

// returns tne next power of 2 > len
int setDim(int len)
{
	int retval=1;
	while(retval < len) retval*=2;
	return retval;
}

// this pads the texture loaded to powers of 2
int load_padded_BMP(char *name, unsigned char **image, int *width, int *height)
{
	FILE *bmpFile;
	BITMAPFILEHEADER bmpFileHeader;
	BITMAPINFOHEADER bmpInfoHeader;
	struct RGBTRIPLE rgb;
	unsigned long i, j, offset;
	unsigned short type;
	int padding, padded_width, padded_height;
	unsigned char scratch[10];

	bmpFile = fopen(name, "rb");
	if (!bmpFile)
		return -1;

	/* first read the first two bytes (type) */
	fread(&type, 2, 1, bmpFile);

	fread(&bmpFileHeader, sizeof(bmpFileHeader), 1, bmpFile); /* Read File Header */
	fread(&bmpInfoHeader, sizeof(bmpInfoHeader), 1, bmpFile); /* Read Info Header */
    padded_width = setDim(bmpInfoHeader.biWidth);   
    padded_height = setDim(bmpInfoHeader.biHeight);   
    if (bmpInfoHeader.biCompression)
		return -2; /* compression used */

    if (bmpInfoHeader.biBitCount != 24)
		return -3; /* not 24 bpp */

    *image = (unsigned char *)malloc(padded_width*padded_height*4);
    *height = padded_height; //bmpInfoHeader.biHeight;
    *width = padded_width; //bmpInfoHeader.biWidth;
    offset=0;

	/* number of bytes in each line of a bmp file is a multiple of 4.
	 * provide adequate padding
	 */
	padding = (4 - (*width * 3) % 4) % 4;


    for (j=0; j<padded_height; j++) {
		for (i=0; i<padded_width; i++) {
			if(j<bmpInfoHeader.biHeight && i<bmpInfoHeader.biWidth)
				fread(&rgb, sizeof(rgb), 1, bmpFile);
			else // pad with blanks
				rgb.rgbtRed = rgb.rgbtGreen = rgb.rgbtBlue = (unsigned char) 0;
			
			(*image)[offset] = rgb.rgbtRed;
	        	offset++;

		    	(*image)[offset] = rgb.rgbtGreen;
	        	offset++;

		    	(*image)[offset] = rgb.rgbtBlue;
			offset++;

			(*image)[offset] = 255;
		    	offset++;
		}
		fread(scratch, padding, 1, bmpFile); /* read the padding and ignore it */
	}

	fclose(bmpFile);

	return 0;
}
